"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const turf_1 = require("@turf/turf");
// @ts-ignore
const parser_1 = require("./parser"); // eslint-disable-line import/no-unresolved
const typeOf = (v) => {
    const t = typeof v;
    if (t !== "object")
        return t;
    if (v === null)
        return "null";
    return Array.isArray(v) ? "array" : t;
};
const def = (name, argTypes, f) => {
    const lastType = argTypes[argTypes.length - 1];
    const variableType = lastType && lastType.variable ? lastType : null;
    const types = (variableType ? argTypes.slice(0, -1) : argTypes).map(t => typeof t === "string" ? { type: t, optional: false } : t);
    const requiredTypes = types.filter(t => !t.optional);
    const isVariable = variableType || requiredTypes.length !== types.length;
    return function fn(...a) {
        if (isVariable
            ? a.length < requiredTypes.length
            : a.length !== requiredTypes.length) {
            throw new parser_1.SyntaxError(`The ${name} function requires ${isVariable ? "at least " : ""}${requiredTypes.length} argument(s)`);
        }
        if (!types.every((t, i) => t.type === "any" ||
            typeOf(a[i]) === t.type ||
            (typeOf(a[i]) === "undefined" && t.optional))) {
            return undefined;
        }
        if (variableType &&
            variableType.type !== "any" &&
            !a.slice(types.length).every(v => typeOf(v) === variableType.type)) {
            return undefined;
        }
        return f(...a);
    };
};
const deepEqual = (a, b, partial) => {
    const typeOfA = typeOf(a);
    const typeOfB = typeOf(b);
    if (typeOfA === "array" && typeOfB === "array") {
        if (a.length !== b.length)
            return false;
        return a.every((v, i) => deepEqual(v, b[i], partial));
    }
    if (typeOfA === "object" && typeOfB === "object") {
        const aEntries = Object.entries(a);
        const bEntries = Object.entries(b);
        if (!partial && aEntries.length !== bEntries.length)
            return false;
        return bEntries.every(([k, v]) => deepEqual(a[k], v, partial));
    }
    return a === b;
};
exports.ABS = def("ABS", ["number"], (v) => Math.abs(v));
exports.ACOS = def("ACOS", ["number"], (v) => Math.acos(v));
exports.ARRAY_CONCAT = def("ARRAY_CONCAT", ["array", "array", { type: "array", variable: true }], (...a) => a.reduce((b, c) => [...b, ...c], []));
exports.ARRAY_CONTAINS = def("ARRAY_CONTAINS", ["array", "any", { type: "boolean", optional: true }], (a, c, partial = false) => a.some(v => deepEqual(v, c, partial)));
exports.ARRAY_LENGTH = def("ARRAY_LENGTH", ["array"], (a) => a.length);
exports.ARRAY_SLICE = def("ARRAY_SLICE", ["array", "number", { type: "number", optional: true }], (a, b, c) => a.slice(b, c != null ? b + c : undefined));
exports.ASIN = def("ASIN", ["number"], (v) => Math.asin(v));
exports.ATAN = def("ATAN", ["number"], (v) => Math.atan(v));
exports.ATN2 = def("ASN2", ["number", "number"], (a, b) => Math.atan2(b, a));
exports.CEILING = def("CEILING", ["number"], (v) => Math.ceil(v));
exports.CONCAT = def("CONCAT", ["string", "string", { type: "string", variable: true }], (...a) => a.join(""));
exports.CONTAINS = def("CONTAINS", ["string", "string"], (a, b) => a.includes(b));
exports.COS = def("COS", ["number"], (v) => Math.cos(v));
exports.COT = def("COT", ["number"], (v) => 1 / Math.tan(v));
exports.DEGREES = def("DEGREES", ["number"], (v) => (v * 180) / Math.PI);
exports.ENDSWITH = def("ENDSWITH", ["string", "string"], (a, b) => a.endsWith(b));
exports.EXP = def("EXP", ["number"], (v) => Math.exp(v));
exports.FLOOR = def("FLOOR", ["number"], (v) => Math.floor(v));
exports.INDEX_OF = def("INDEX_OF", ["string", "string"], (a, b) => a.indexOf(b));
exports.IS_ARRAY = def("IS_ARRAY", ["any"], (v) => Array.isArray(v));
exports.IS_BOOL = def("IS_BOOL", ["any"], (v) => typeof v === "boolean");
exports.IS_DEFINED = def("IS_DEFINED", ["any"], (v) => typeof v !== "undefined");
exports.IS_NULL = def("IS_NULL", ["any"], (v) => v === null);
exports.IS_NUMBER = def("IS_NUMBER", ["any"], (v) => typeof v === "number");
exports.IS_OBJECT = def("IS_OBJECT", ["any"], (v) => Boolean(v) && typeof v === "object" && !Array.isArray(v));
exports.IS_STRING = def("IS_STRING", ["any"], (v) => typeof v === "string");
exports.IS_PRIMITIVE = def("IS_PRIMITIVE", ["any"], (v) => exports.IS_NULL(v) || exports.IS_NUMBER(v) || exports.IS_STRING(v) || exports.IS_BOOL(v));
exports.LEFT = def("LEFT", ["string", "number"], (a, b) => a.slice(0, b));
exports.LENGTH = def("LENGTH", ["string"], (v) => v.length);
exports.LOG = def("LOG", ["number"], (v) => Math.log(v));
exports.LOG10 = def("LOG10", ["number"], (v) => Math.log10(v));
exports.LOWER = def("LOWER", ["string"], (v) => v.toLowerCase());
exports.LTRIM = def("LTRIM", ["string"], (v) => v.trimLeft());
exports.PI = def("PI", [], () => Math.PI);
exports.POWER = def("POWER", ["number", "number"], (a, b) => a ** b);
exports.RADIANS = def("RADIANS", ["number"], (v) => (v * Math.PI) / 180);
exports.REPLACE = def("REPLACE", ["string", "string", "string"], (a, b, c) => a.replace(b, c));
exports.REPLICATE = def("REPLICATE", ["string", "number"], (a, b) => a.repeat(b));
exports.REVERSE = def("REVERSE", ["string"], (v) => v
    .split("")
    .reverse()
    .join(""));
exports.RIGHT = def("RIGHT", ["string", "number"], (a, b) => a.slice(-b));
exports.ROUND = def("ROUND", ["number"], (v) => Math.round(v));
exports.RTRIM = def("RTRIM", ["string"], (v) => v.trimRight());
exports.SIGN = def("SIGN", ["number"], (v) => Math.sign(v));
exports.SIN = def("SIN", ["number"], (v) => Math.sin(v));
exports.SQRT = def("SQRT", ["number"], (v) => Math.sqrt(v));
exports.SQUARE = def("SQUARE", ["number"], (v) => v ** 2);
exports.STARTSWITH = def("STARTSWITH", ["string", "string"], (a, b) => a.startsWith(b));
exports.SUBSTRING = def("SUBSTRING", ["string", "number", { type: "number", optional: true }], (a, b, c) => a.substring(b, c != null ? b + c : undefined));
exports.TAN = def("TAN", ["number"], (v) => Math.tan(v));
exports.TOSTRING = def("ToString", ["any"], (v) => {
    const t = typeOf(v);
    if (t === "undefined") {
        return undefined;
    }
    if (t === "object" || t === "array") {
        return JSON.stringify(v);
    }
    return String(v);
});
exports.TRIM = def("TRIM", ["string"], (v) => v.trim());
exports.TRUNC = def("TRUNC", ["number"], (v) => Math.trunc(v));
exports.UPPER = def("UPPER", ["string"], (v) => v.toUpperCase());
// Spatial functions
const spatialBinaryOp = (name, f) => {
    return def(name, ["any", "any"], (a, b) => {
        const t1 = typeOf(a);
        if (t1 === "undefined") {
            return undefined;
        }
        const aObj = t1 === "object" ? a : JSON.parse(a);
        const t2 = typeOf(a);
        if (t2 === "undefined") {
            return undefined;
        }
        const bObj = t1 === "object" ? b : JSON.parse(b);
        const aFeat = turf_1.feature(aObj);
        const bFeat = turf_1.feature(bObj);
        return f(aFeat, bFeat);
    });
};
exports.ST_DISTANCE = spatialBinaryOp("ST_DISTANCE", (a, b) => {
    // Turf can only handle point distances - take the centroid
    // as a workaround.
    const pa = turf_1.centroid(a);
    const pb = turf_1.centroid(b);
    // Convert kilometers to meters.
    return turf_1.distance(pa.geometry.coordinates, pb.geometry.coordinates) * 1000;
});
exports.ST_WITHIN = spatialBinaryOp("ST_WITHIN", turf_1.booleanWithin);
exports.ST_INTERSECTS = spatialBinaryOp("ST_INTERSECTS", (a, b) => !turf_1.booleanDisjoint(a, b));
exports.REGEXMATCH = def("REGEXMATCH", ["string", "string", { type: "string", optional: true }], (str, p, f) => {
    let pattern = p;
    let flags = f;
    if (flags) {
        if (!/^[msix]+$/.test(flags)) {
            throw new Error(`Unexpectd flags on REGEXMATCH: ${flags}`);
        }
        if (flags.includes("x")) {
            pattern = pattern.replace(/\s/g, "");
            flags = flags.replace(/x/g, "");
        }
    }
    // TODO: cache RegExp instances?
    const re = new RegExp(pattern, flags);
    return re.test(str);
});
